#!/usr/bin/env python

"""
rm.py - Rate Monotonic scheduler

RmPriorityQueue: priority queue that prioritizes by period
RmScheduler: scheduling algorithm that executes RM
"""

import json
import sys

from taskset import *
from scheduleralgorithm import *
from schedule import ScheduleInterval, Schedule
from display import SchedulingDisplay

class RmPriorityQueue(PriorityQueue):
    def __init__(self, jobReleaseDict):
        """
        Creates a priority queue of jobs ordered by period.
        """
        PriorityQueue.__init__(self, jobReleaseDict)

    def _sortQueue(self):
        # RM orders by period
        # TODO: implement this!
        raise NotImplementedError

    def _findFirst(self, t):
        """
        Returns the index of the highest-priority job released at or before t,
        or -1 if the queue is empty or if all remaining jobs are released after t.
        """
        # TODO: implement this!
        raise NotImplementedError

    def popNextJob(self, t):
        """
        Removes and returns the highest-priority job of those released at or after t,
        or None if no jobs are released at or after t.
        """
        # TODO: implement this!
        raise NotImplementedError

    def popPreemptingJob(self, t, job):
        """
        Removes and returns the job that will preempt job 'job' after time 't', or None
        if no such preemption will occur (i.e., if no higher-priority jobs
        are released before job 'job' will finish executing).

        t: the time after which a preemption may occur
        job: the job that is executing at time 't', and which may be preempted
        """
        # TODO: implement this!
        raise NotImplementedError

class RmScheduler(SchedulerAlgorithm):
    def __init__(self, taskSet):
        SchedulerAlgorithm.__init__(self, taskSet)
        SchedulerAlgorithm.__init__(self, taskSet)

    def buildSchedule(self, startTime, endTime):
        self._buildPriorityQueue(RmPriorityQueue)

        # TODO: build a scheduling using RM, using _makeSchedulingDecision() to determine
        # the next decision to be made (think about how this differs from FIFO/EDF)

        return self.schedule

    def _makeSchedulingDecision(self, t, previousJob):
        """
        Makes a scheduling decision after time t.

        t: the beginning of the previous time interval, if one exists (or 0 otherwise)
        previousJob: the job that was previously executing, and will either complete or be preempted

        returns: (ScheduleInterval instance, Job instance of new job to execute)
        """
        # TODO: make a scheduling decision at the next scheduling time point after time t.
        # Let's call this new time nextTime.
        #
        # If there was no previous job, be sure to choose the next job from the priority queue at or after time t.
        # If there was a previous job, choose the highest-priority job from the priority queue of those released at or before time nextTime.
        # Make sure to handle preemptions properly!
        #
        # Note that if there was a previous job but when it finishes, no more jobs are ready (released prior to that time),
        # then there should be no job associated with the new interval.  The ScheduleInterval constructor will handle this, you
        # just have to provide it with None rather than a job.
        #
        # Once you have the new job to execute (if any), build the interval (which starts at nextTime) and return it and the new job.

        return interval, newJob

if __name__ == "__main__":
    if len(sys.argv) > 1:
        file_path = sys.argv[1]
    else:
        file_path = "tasksets/hwp1_test1.json"

    with open(file_path) as json_data:
        data = json.load(json_data)

    taskSet = TaskSet(data)

    taskSet.printTasks()
    taskSet.printJobs()

    rm = RmScheduler(taskSet)
    schedule = rm.buildSchedule(0, 6)

    schedule.printIntervals(displayIdle=True)

    print("\n// Validating the schedule:")
    schedule.checkWcets()
    schedule.checkFeasibility()

    display = SchedulingDisplay(width=800, height=480, fps=33, scheduleData=schedule)
    display.run()
